STUDIO_CONFIG_KEY = "studio"

KEYCLOAK_ACCESS_TOKEN_KEY = "access_token"
KEYCLOAK_EXPIRES_IN_KEY = "expires_in"
KEYCLOAK_REFRESH_EXPIRES_IN_KEY = "refresh_expires_in"
KEYCLOAK_REFRESH_TOKEN = "refresh_token"
KEYCLOAK_TOKEN_TYPE = "token_type"

RASA_STUDIO_AUTH_SERVER_URL_ENV = "RASA_STUDIO_AUTH_SERVER_URL"
RASA_STUDIO_CLI_STUDIO_URL_ENV = "RASA_STUDIO_CLI_STUDIO_URL"
RASA_STUDIO_CLI_REALM_NAME_KEY_ENV = "RASA_STUDIO_CLI_REALM_NAME_KEY"
RASA_STUDIO_CLI_CLIENT_ID_KEY_ENV = "RASA_STUDIO_CLI_CLIENT_ID_KEY"
RASA_STUDIO_CLI_DISABLE_VERIFY_KEY_ENV = "RASA_STUDIO_CLI_DISABLE_VERIFY_KEY"

STUDIO_NLU_FILENAME = "studio_nlu.yml"
STUDIO_DOMAIN_FILENAME = "studio_domain.yml"
DOMAIN_FILENAME = "domain.yml"
STUDIO_FLOWS_FILENAME = "studio_flows.yml"
STUDIO_CONFIG_FILENAME = "studio_config.yml"
STUDIO_ENDPOINTS_FILENAME = "studio_endpoints.yml"
